/*
 *  biquad.h
   
	(c) Risto Holopainen 2014
	
	Copying and distribution of this file, with or without modification,
	are permitted in any medium without royalty provided the copyright
	notice and this notice are preserved.  This file is offered as-is,
	without any warranty.	

	Created by Risto Holopainen in 2006--2012.
 */



#ifndef BIQUADS_H
#define BIQUADS_H

#include <cmath>

/*
	first, a few simple things:
	one_pole
	LP3			--- simplest FIR lowpass
	allpass1p	--- first order allpass

	then the biquads proper:
	dc_block : remove DC (and Nyquist components)
	lopass
	hipass 
	bandpass
	allpass ( with bandpass and band reject)
*/


// 1-pole LP/HP
template <typename T>
class one_pole
{
	public:
		one_pole(double a0=1.0, double b1=0.75) {	y = y1 = T(0.0);	a = a0; b = b1;	}
		T filter(T x);

	private:
		T y, y1;
		double a, b;
};

template <typename T>
T one_pole<T>::filter(T x)
{
	y = a*x + b*y1;
	y1 = y;
	return y;
}


template <typename T>
class LP3
{
	public:
		LP3() : x1(0), x2(0) {;}
		T filter(T);
	private:
		T x1, x2, y;
};

template <typename T>
T LP3<T>::filter(T x)
{
	y =	.25*(x + 2*x1 + x2);
	x2 = x1; x1 = x;
	return y;
}

class allpass1p
{ 
// ref. Zölzer's DAFX

	public:	
		allpass1p(int sr, float fc);
		void cfr(float f);
		double filter(double);
		
	private:	
		double c, omg;
		double xh0, xh1, y;
};

allpass1p::allpass1p(int sr, float fc)
{

	omg = M_PI/sr;
	double w = tan(omg*fc);
	c = (w - 1) / (w + 1);
	xh0 = xh1 = 0;
}

double allpass1p::filter(double x)
{
	xh0 = x - c*xh1;
	y = c*xh0 + xh1;
	xh1 = xh0;
	return y;
}

void allpass1p::cfr(float fc)
{
	double w = tan(omg*fc);
	c = (w - 1) / (w + 1);
}


class dc_block
{
// has zeros at DC and nyquist.
	public:
		dc_block(double b=0.995) 
		{B = (b > 0 && b < 1.0) ? b : 0.995; y1 = y2 = 0; x1 = x2 = 0; }
		double filter(const double);

	private:
		double B;
		double y, y1, y2, x1, x2;
};


double dc_block::filter(const double x)
{
	y = x - x2 + B*y2;
	x2 = x1; x1 = x; y2 = y1; y1 = y;
	return y;	
}


//------------------------------------
 
class lopass
{
// ref. Zölzer DAFX
	public:
		lopass(const int sr, const float fc);
		double filter(const double);
		void cutoff(const float);

	private:
		int sr;
		double K, s2;
		double a0, a1, a2, b1, b2;
		double x1, x2, y, y1, y2;
};

lopass::lopass(const int fs, const float fc)
{
	sr = fs;
	s2 = sqrt(2.0);
	cutoff(fc);
	x1=x2=y1=y2=0;
}

double lopass::filter(const double x)
{
	y = a0*x + a1*x1 + a2*x2 - b1*y1 - b2*y2;	
	x2 = x1; x1 = x;
	y2 = y1; y1 = y;
	return y;
}

void lopass::cutoff(const float fc)
{
	
	K = tan(M_PI*fc/sr);
	double K2 = K*K;
	a0 = K2/(1.0 + s2*K + K2);
	a1 = 2.0 * a0;
	a2 = a0;
	b1 = 2.0*(K2 -1)/(1.0 + s2*K + K2);
	b2 = (1.0 - s2*K + K2)/(1.0 + s2*K + K2); 
}


//------------------------------------------


class hipass
{
// ref. Zölzer DAFX
	public:
		hipass(const int sr, const float fc);
		double filter(const double);
		void cutoff(const float);

	private:
		int sr;
		double K, s2;
		double a0, a1, a2, b1, b2;
		double x1, x2, y, y1, y2;
};

hipass::hipass(const int fs, const float fc)
{
	sr = fs;
	s2 = sqrt(2.0);
	cutoff(fc);
	x1=x2=y1=y2=0;
}

double hipass::filter(const double x)
{
	y = a0*x + a1*x1 + a2*x2 - b1*y1 - b2*y2;
	x2 = x1; x1 = x;
	y2 = y1; y1 = y;
	return y;
}

void hipass::cutoff(const float fc)
{
	K = tan(M_PI*fc/sr);
	double K2 = K*K;
	a0 = 1.0/(1.0 + s2 + K2);
	a1 = -2.0/(1.0 + s2 + K2);
	a2 = 1.0/(1.0 + s2 + K2);
	b1 = 2.0*(K2 -1)/(1.0 + s2*K + K2);	// samma b1, b2 som i lopass
	b2 = (1.0 - s2*K + K2)/(1.0 + s2*K + K2); 
}

//---------------------------------------------------



class bandpass
{
/*
	After recipe in Moore, Elements of Computer Music
*/

	public:
		bandpass(double sr, double frq=1000, double q=1.0);
		double filter(double);
		void cfr(float f);		// set new centre frequency. 

	private:
		double sr;
		double frq;
		double Q;
		double x[3], y[3];
		double R;
		double G;
		double b1, b2;
		double twopiosr;
};

bandpass::bandpass(double f_sr, double F, double q)
{
	sr = f_sr;
	frq = F;
	Q = q;

	double BW = frq/Q;
	R = exp(-M_PI*BW/sr);
	G = 1.0 - R;
	twopiosr = 2*M_PI/sr;
	double theta = twopiosr*frq;

	b1 = 2*R*cos(theta);
	b2 = -R*R;

	x[0] = x[1] = x[2] = 0;
	y[0] = y[1] = y[2] = 0;
}


double bandpass::filter(const double x0)
{
	x[0] = x0;
	y[0] = G*(x[0] - R*x[2]) + b1*y[1] + b2*y[2];
	for(int n=2; n>0; n--)
		{
		x[n] = x[n-1];
		y[n] = y[n-1];
		}
	return y[0];
}

void bandpass::cfr(const float frq)
{
	double BW = frq/Q;
	R = exp(-M_PI*BW/sr);
	G = 1.0 - R;
	b1 = 2*R*cos(twopiosr*frq);
	b2 = -R*R;	
}

/*--------------------------------------------
	2nd order allpass
	this filter can also be used for bandpass or band reject
*/
typedef double fp_type;

template <typename T> // float, double
class allpass
{
	public: // sr, centerfrq, bandwidth hz
		allpass(int=44100, fp_type fc=440, fp_type fb=40);
		void cfr(fp_type CF, fp_type BW); 	// ny centerfrekv och bandbredd 
		T filter(const T);
		T bandpass(const T); 
		T bandreject(const T); 
  
	private:
		int sr;
		T x[3],  y[3];
		fp_type c, d; // internal variables
		fp_type a;
};

template <typename T>
allpass<T>::allpass(int s, fp_type fc, fp_type fb)
{
	sr = s;
	this->cfr(fc, fb);
	x[0] = x[1] = x[2] = y[0] = y[1] = y[2] = 0;
}

template <typename T>
void allpass<T>::cfr(fp_type fc, fp_type fb)
{
	fp_type r = tan(M_PI*fb/sr);
	c = (r-1)/(r+1);
	d = -cos(2*M_PI*fc/sr);
	a = d*(1-c);
}

template <typename T>
T allpass<T>::filter(const T xn)
{
	x[0] = xn;
//	y[0] = -c*x[0] + a*x[1] + x[2] -a*y[1] + c*y[2];
	y[0] = x[2] + c*(y[2] - x[0]) + a*(x[1] - y[1]);

	for(int i=2; i>0; i--)
  		{
  		x[i] = x[i-1];
  		y[i] = y[i-1];
   		}
	return y[0];
}


template <typename T>
T allpass<T>::bandpass(const T xn)
{
	return 0.5*(xn - this->filter(xn)); 
}



template <typename T>
T allpass<T>::bandreject(const T xn)
{
	return 0.5*(xn + this->filter(xn)); 
}




class notch
{
// biquad notch. centre frq and pole radius < 1.
	public:
		notch(int sr, double cfr, double r);
		double filter(double x);
		void cfr(double cf);
		void pole(double pp); 
		
	private:
		int sr;
		double Omega;
		double w0;	// centre frq in radians
		double r;	// pole radius
		double x[2], y[3];
		double a, b1, b2;
};



notch::notch(int fs, double cf, double pr)
{
	sr = fs;
	Omega = 2.0*M_PI/sr;
	w0 = Omega*cf;
	r = (pr > 0 && pr < 1) ? pr : 0.95;
	a = 2*cos(w0);
	b1 = 2*r*cos(w0);
	b2 = r*r;
}


double notch::filter(double x_in)
{
	y[0] = x_in - a*x[0] + x[1] + b1*y[1] - b2*y[2];
	
	x[1] = x[0];
	x[0] = x_in;
	y[2] = y[1];
	y[1] = y[0];
	return y[0];
}

void notch::cfr(double cf)
{ 
	w0 = Omega*cf; 
	a = 2.0*cos(w0);
	b1 = 2*r*cos(w0);
}



void notch::pole(double pp)
{
	r = pp;
	b1 = 2.0*r*cos(w0);
	b2 = r*r;
}



// ----------------------------------------





#endif
